"use strict";
(() => {
  // src/helpers.ts
  function padLeftTo32Bytes(input) {
    const length = input.length % 32 ? input.length + 32 - input.length % 32 : input.length;
    const result = new Uint8Array(length);
    result.set(input, result.length - input.length);
    return result;
  }
  function padRightTo32Bytes(input) {
    const length = input.length % 32 ? input.length + 32 - input.length % 32 : input.length;
    const result = new Uint8Array(length);
    result.set(input, 0);
    return result;
  }
  function concatenateBytes(source) {
    return new Uint8Array(source.flatMap((x) => [...x]));
  }
  function padAndLengthPrefix(source) {
    const length = source.length;
    const padded = padRightTo32Bytes(source);
    return concatenateBytes([integerToBytes(length), padded]);
  }
  function encodeDynamicData(encodedData) {
    let staticBytesSize = 0;
    for (let encodedParameter of encodedData) {
      if (encodedParameter.isDynamic) staticBytesSize += 32;
      else staticBytesSize += encodedParameter.bytes.length;
    }
    const staticBytes = [];
    const dynamicBytes = [];
    for (let encodedParameter of encodedData) {
      if (encodedParameter.isDynamic) {
        const dynamicBytesAppendedSoFar = dynamicBytes.reduce(
          (total, bytes) => total += bytes.length,
          0
        );
        staticBytes.push(integerToBytes(staticBytesSize + dynamicBytesAppendedSoFar));
        dynamicBytes.push(encodedParameter.bytes);
      } else {
        staticBytes.push(encodedParameter.bytes);
      }
    }
    return concatenateBytes([...staticBytes, ...dynamicBytes]);
  }
  function isEncodableArray(maybe) {
    return Array.isArray(maybe);
  }
  function integerToBytes(value, byteWidth = 32, signed = false) {
    return signed ? signedToBytes(value, byteWidth) : unsignedToBytes(value, byteWidth);
  }
  function unsignedToBytes(value, byteWidth = 32) {
    if (typeof value === "number") value = BigInt(value);
    const bits = byteWidth * 8;
    if (value >= 2n ** BigInt(bits) || value < 0n)
      throw new Error(`Cannot fit ${value} into a ${bits}-bit unsigned integer.`);
    const result = new Uint8Array(byteWidth);
    for (let i = 0; i < byteWidth; ++i) {
      result[i] = Number(value >> BigInt(bits - i * 8 - 8) & 0xffn);
    }
    return result;
  }
  function signedToBytes(value, byteWidth = 32) {
    if (typeof value === "number") value = BigInt(value);
    const bits = byteWidth * 8;
    if (value >= 2n ** (BigInt(bits) - 1n) || value < -(2n ** (BigInt(bits) - 1n)))
      throw new Error(`Cannot fit ${value} into a ${bits}-bit signed integer.`);
    const unsignedValue = twosComplement(value, bits);
    return unsignedToBytes(unsignedValue);
  }
  function twosComplement(value, numberOfBits) {
    const mask = 2n ** (BigInt(numberOfBits) - 1n) - 1n;
    return (value & mask) - (value & ~mask);
  }

  // src/calldata.ts
  function parseSignature(functionSignature) {
    const signatureMatcher = /^([a-zA-Z_][a-zA-Z0-9_]+)\((.*)\)$/;
    const matchedSignature = signatureMatcher.exec(functionSignature);
    if (matchedSignature === null)
      throw new Error(`${functionSignature} is not a valid Solidity function signature.`);
    const name = matchedSignature[1];
    const inputs = parseParameters(matchedSignature[2]);
    return { type: "function", name, inputs, outputs: [] };
  }
  function parseParameters(functionParameters) {
    const parameters = [];
    let remainingParameters = functionParameters.trim();
    while (remainingParameters.length !== 0) {
      let { parameterDescription, remaining } = extractNextParameter(remainingParameters);
      remainingParameters = remaining;
      parameters.push(parameterDescription);
    }
    return parameters.map((x, i) => ({ ...x, name: x.name || `arg${i}` }));
  }
  function extractNextParameter(functionParameters) {
    let nesting = 0;
    let typeAndName = "";
    for (const character of functionParameters) {
      if (character === "(") ++nesting;
      if (character === ")") --nesting;
      if (nesting < 0)
        throw new Error(
          `${functionParameters} does not have matching number of open and close parenthesis`
        );
      if (nesting > 0) {
        typeAndName += character;
        continue;
      }
      if (character === ",") break;
      typeAndName += character;
    }
    const typeAndNameMatch = /^\s*(.+?)\s*(?:\s([a-zA-Z_][a-zA-Z0-9_]*))?\s*$/.exec(typeAndName);
    if (typeAndNameMatch === null)
      throw new Error(`${typeAndNameMatch} is not a valid parameter/name pair.`);
    let parameterType = typeAndNameMatch[1];
    let components = void 0;
    if (parameterType.startsWith("(")) {
      const tupleTypes = parameterType.slice(1, parameterType.lastIndexOf(")"));
      parameterType = `tuple${parameterType.slice(tupleTypes.length + 2)}`;
      components = parseParameters(tupleTypes);
    }
    const parameterName = typeAndNameMatch[2] || "";
    let remaining = functionParameters.slice(typeAndName.length);
    if (remaining.startsWith(",")) remaining = remaining.slice(1);
    remaining = remaining.trim();
    const parameterDescription = {
      name: parameterName,
      type: parameterType,
      components
    };
    return { parameterDescription, remaining };
  }
  function generateFullSignature(functionDescription) {
    return `${functionDescription.name}(${toFullParameters(functionDescription.inputs)})`;
  }
  function generateCanonicalSignature(functionDescription) {
    return `${functionDescription.name}(${toCanonicalParameters(functionDescription.inputs)})`;
  }
  function toFullParameters(parameters) {
    return parameters.map(toFullParameter).join(", ");
  }
  function toCanonicalParameters(parameters) {
    return parameters.map(toCanonicalParameter).join(",");
  }
  function toFullParameter(parameter) {
    if (parameter.type.startsWith("tuple")) {
      if (parameter.components === void 0)
        throw new Error(
          `Encountered a 'tuple' type that had no components.  Did you mean to include an empty array?`
        );
      return `(${toFullParameters(parameter.components)})${parameter.type.slice("tuple".length)} ${parameter.name}`;
    } else {
      return `${parameter.type} ${parameter.name}`;
    }
  }
  function toCanonicalParameter(parameter) {
    if (parameter.type.startsWith("tuple")) {
      if (parameter.components === void 0)
        throw new Error(
          `Encountered a 'tuple' type that had no components.  Did you mean to include an empty array?`
        );
      return `(${toCanonicalParameters(parameter.components)})${parameter.type.slice("tuple".length)}`;
    } else {
      return parameter.type;
    }
  }
  function encodeMethod(first, second, parameters) {
    if (typeof first === "number")
      return encodeMethodWithSelector(
        first,
        second,
        parameters
      );
    else if (typeof second === "string") return encodeMethodWithSignature(first, second, parameters);
    else return encodeMethodWithDescription(first, second, parameters);
  }
  async function encodeMethodWithDescription(keccak256, functionDescription, parameters) {
    const canonicalSignature = generateCanonicalSignature(functionDescription);
    const canonicalSignatureHash = await keccak256(new TextEncoder().encode(canonicalSignature));
    const functionSelector = canonicalSignatureHash >> 224n;
    return encodeMethod(Number(functionSelector), functionDescription.inputs, parameters);
  }
  async function encodeMethodWithSignature(keccak256, functionSignature, parameters) {
    const functionDescription = parseSignature(functionSignature);
    return await encodeMethodWithDescription(keccak256, functionDescription, parameters);
  }
  function encodeMethodWithSelector(functionSelector, parameterDescriptions, parameters) {
    const encodedParameters = encodeParameters(parameterDescriptions, parameters);
    return new Uint8Array([...integerToBytes(functionSelector, 4), ...encodedParameters]);
  }
  function encodeParameters(descriptions, parameters) {
    if (descriptions.length !== parameters.length)
      throw new Error(
        `Number of provided parameters (${parameters.length}) does not match number of expected parameters (${descriptions.length})`
      );
    const encodedParameters = parameters.map(
      (nestedParameter, index) => encodeParameter(descriptions[index], nestedParameter)
    );
    return encodeDynamicData(encodedParameters);
  }
  function encodeParameter(description, parameter) {
    return tryEncodeFixedArray(description, parameter) || tryEncodeDynamicArray(description, parameter) || tryEncodeTuple(description, parameter) || tryEncodeDynamicBytes(description, parameter) || tryEncodeString(description, parameter) || tryEncodeBoolean(description, parameter) || tryEncodeNumber(description, parameter) || tryEncodeAddress(description, parameter) || tryEncodeFixedBytes(description, parameter) || tryEncodeFixedPointNumber(description) || tryEncodeFunction(description) || function() {
      throw new Error(`Unsupported parameter type ${description.type}`);
    }();
  }
  function tryEncodeFixedArray(description, parameter) {
    const match = /^(.*)\[(\d+)\]$/.exec(description.type);
    if (match === null) return null;
    const size = Number.parseInt(match[2]);
    if (!Array.isArray(parameter) || parameter.length !== size)
      throw new Error(
        `Can only encode a JavaScript 'array' of length ${size} into an EVM 'array' of length ${size}
${parameter}`
      );
    const nestedDescription = Object.assign({}, description, { type: match[1] });
    const encodedParameters = parameter.map(
      (nestedParameter) => encodeParameter(nestedDescription, nestedParameter)
    );
    const isDynamic = encodedParameters.some((x) => x.isDynamic);
    if (isDynamic) {
      return { isDynamic, bytes: encodeDynamicData(encodedParameters) };
    } else {
      return { isDynamic, bytes: concatenateBytes(encodedParameters.map((x) => x.bytes)) };
    }
  }
  function tryEncodeDynamicArray(description, parameter) {
    if (!description.type.endsWith("[]")) return null;
    if (!Array.isArray(parameter))
      throw new Error(`Can only encode a JavaScript 'array' into an EVM 'array'
${parameter}`);
    const nestedDescription = Object.assign({}, description, {
      type: description.type.substring(0, description.type.length - 2)
    });
    const encodedParameters = parameter.map(
      (nestedParameter) => encodeParameter(nestedDescription, nestedParameter)
    );
    const lengthBytes = integerToBytes(encodedParameters.length);
    return {
      isDynamic: true,
      bytes: concatenateBytes([lengthBytes, encodeDynamicData(encodedParameters)])
    };
  }
  function tryEncodeTuple(description, parameter) {
    if (description.type !== "tuple") return null;
    if (typeof parameter !== "object")
      throw new Error(
        `Can only encode a JavaScript 'object' or a JavaScript array into an EVM 'tuple'
${parameter}`
      );
    if (description.components === void 0 || description.components.length === 0) {
      return { isDynamic: false, bytes: new Uint8Array(0) };
    } else {
      const encodableTupleOrArray = parameter;
      const encodedComponents = description.components.map((component, index) => {
        const parameter2 = isEncodableArray(encodableTupleOrArray) ? encodableTupleOrArray[index] : encodableTupleOrArray[component.name];
        return encodeParameter(component, parameter2);
      });
      const isDynamic = encodedComponents.some((x) => x.isDynamic);
      return {
        isDynamic,
        bytes: isDynamic ? encodeDynamicData(encodedComponents) : concatenateBytes(encodedComponents.map((x) => x.bytes))
      };
    }
  }
  function tryEncodeDynamicBytes(description, parameter) {
    if (description.type !== "bytes") return null;
    if (!(parameter instanceof Uint8Array))
      throw new Error(`Can only encode a JavaScript 'Uint8Array' into EVM 'bytes'
${parameter}`);
    return { isDynamic: true, bytes: padAndLengthPrefix(parameter) };
  }
  function tryEncodeString(description, parameter) {
    if (description.type !== "string") return null;
    if (typeof parameter !== "string")
      throw new Error(`Can only encode a JavaScript 'string' into an EVM 'string'
${parameter}`);
    const encoded = new TextEncoder().encode(parameter);
    return { isDynamic: true, bytes: padAndLengthPrefix(encoded) };
  }
  function tryEncodeBoolean(description, parameter) {
    if (description.type !== "bool") return null;
    if (typeof parameter !== "boolean")
      throw new Error(`Can only encode JavaScript 'boolean' into EVM 'bool'
${parameter}`);
    const bytes = new Uint8Array(32);
    bytes.set([parameter ? 1 : 0], 31);
    return { isDynamic: false, bytes };
  }
  function tryEncodeNumber(description, parameter) {
    const match = /^(u?)int(\d*)$/.exec(description.type);
    if (match === null) return null;
    if (typeof parameter !== "bigint")
      throw new Error(
        `Can only encode a JavaScript 'bigint' into an EVM '${description.type}'
${parameter}`
      );
    const size = Number.parseInt(match[2]);
    if (size <= 0 || size > 256 || size % 8)
      throw new Error(`EVM numbers must be in range [8, 256] and must be divisible by 8.`);
    const signed = !match[1];
    if (!signed && parameter >= 2n ** BigInt(size))
      throw new Error(
        `Attempted to encode ${parameter} into a ${description.type}, but it is too big to fit.`
      );
    if (!signed && parameter < 0n)
      throw new Error(
        `Attempted to encode ${parameter} into a ${description.type}, but you cannot encode negative numbers into a ${description.type}.`
      );
    if (signed && parameter >= 2n ** BigInt(size - 1))
      throw new Error(
        `Attempted to encode ${parameter} into a ${description.type}, but it is too big to fit.`
      );
    if (signed && parameter < -(2n ** BigInt(size - 1)))
      throw new Error(
        `Attempted to encode ${parameter} into a ${description.type}, but it is too big (of a negative number) to fit.`
      );
    const bytes = integerToBytes(parameter, 32, signed);
    return { isDynamic: false, bytes };
  }
  function tryEncodeAddress(description, parameter) {
    if (description.type !== "address") return null;
    if (typeof parameter !== "bigint")
      throw new Error(`Can only encode JavaScript 'bigint' into EVM 'address'
${parameter}`);
    if (parameter > 0xffffffffffffffffffffffffffffffffffffffffn)
      throw new Error(
        `Attempted to encode 0x${parameter.toString(16)} into an EVM address, but it is too big to fit.`
      );
    if (parameter < 0n)
      throw new Error(
        `Attempted to encode ${parameter} into an EVM address, but addresses must be positive numbers.`
      );
    return { isDynamic: false, bytes: padLeftTo32Bytes(integerToBytes(parameter, 20)) };
  }
  function tryEncodeFixedBytes(description, parameter) {
    const match = /^bytes(\d+)$/.exec(description.type);
    if (match === null) return null;
    const size = Number.parseInt(match[1]);
    if (typeof parameter !== "bigint")
      throw new Error(`Can only encode JavaScript 'bigint' into EVM 'bytes${size}'
${parameter}`);
    if (parameter >= 2n ** BigInt(size * 8))
      throw new Error(
        `Attempted to encode 0x${parameter.toString(16)} into an EVM ${description.type}, but it is too big to fit.`
      );
    if (parameter < 0n)
      throw new Error(
        `Attempted to encode -0x${parameter.toString(16).slice(1)} into an EVM ${description.type}, but you cannot encode negative numbers into a ${description.type}.`
      );
    return { isDynamic: false, bytes: padRightTo32Bytes(integerToBytes(parameter, size)) };
  }
  function tryEncodeFixedPointNumber(description) {
    if (!/^u?fixed\d+x\d+$/.test(description.type)) return null;
    throw new Error(`Encoding into EVM type ${description.type} is not supported`);
  }
  function tryEncodeFunction(description) {
    if (description.type !== "function") return null;
    throw new Error(`Encoding into EVM type ${description.type} is not supported`);
  }

  // src/multicall.ts
  function prepareMulticallCalldata(calls) {
    const functionDesc = parseSignature("multicall((uint256,uint256,uint256[])[])");
    const encodedParams = wireEncodeByteArray(
      encodeParameters(functionDesc.inputs, [
        calls.map((call) => [
          BigInt(call.contract_address),
          BigInt(call.entry_point),
          [...call.calldata.map((cd) => BigInt(cd))]
        ])
      ])
    );
    return encodedParams.replace("0x", "0x76971d7f");
  }
  function wireEncodeByteArray(bytes) {
    let result = "";
    for (let i = 0; i < bytes.length; ++i) {
      result += ("0" + bytes[i].toString(16)).slice(-2);
    }
    return `0x${result}`;
  }

  // src/types/index.ts
  var types_exports = {};
})();
//# sourceMappingURL=index.js.map